@@ -36,7 +36,7 @@
 #include <gssapi/gssapi.h>
 
 #define MAX_FILE_NAMELEN	32
-#define FD_ALLOC_BLOCK		32
+#define FD_ALLOC_BLOCK		256
 #ifndef GSSD_PIPEFS_DIR
 #define GSSD_PIPEFS_DIR		"/var/lib/nfs/rpc_pipefs"
 #endif
@@ -58,9 +58,12 @@ enum {AUTHTYPE_KRB5, AUTHTYPE_SPKM3, AUTHTYPE_LIPKEY};
 
 
 
-extern char			pipefsdir[PATH_MAX];
+extern char			pipefs_dir[PATH_MAX];
+extern char			pipefs_nfsdir[PATH_MAX];
 extern char			keytabfile[PATH_MAX];
 extern char			ccachedir[PATH_MAX];
+extern int			use_memcache;
+extern int			root_uses_machine_creds;
 
 TAILQ_HEAD(clnt_list_head, clnt_info) clnt_list;
 
